"""
Shell cloud API module, documentation:
https://shelly-api-docs.shelly.cloud/cloud-control-api/communication
"""

from typing import Optional, Union
import requests

from .validations import validate_empty_string


class ShellyApi:
    """
    Class providing methods for getting devices, their status and ids
    via API calls.
    """

    def __init__(self, host: str, token: str, timeout: int = 5) -> None:
        validate_empty_string(host, "host")
        validate_empty_string(token, "token")
        self.host: str = host
        self.token: str = token
        self.timeout: int = timeout

    def __query_status_api(self, path: str, id: Optional[str | int] = None) -> dict:
        try:
            payload = {"auth_key": self.token}

            if id:
                payload["id"] = id

            status_request = requests.post(
                f"{self.host}/device/{path}", data=payload, timeout=self.timeout
            )
            status = status_request.json()
            return status
        except (
            requests.exceptions.ConnectTimeout,
            requests.exceptions.ConnectionError,
        ):
            return {"success": False, "msg": "Request couldn't connect or timed out"}

    def __set_parameter(self, path: str, data: dict) -> dict:
        try:
            payload = {"auth_key": self.token} | data

            status_request = requests.post(
                f"{self.host}/device/{path}",
                data=payload,
                timeout=self.timeout,
            )
            status = status_request.json()
            return status
        except (
            requests.exceptions.ConnectTimeout,
            requests.exceptions.ConnectionError,
        ):
            return {"success": False, "msg": "Request couldn't connect or timed out"}

    def get_all_devices(self) -> dict:
        """Get the status of ALL devices"""
        response = self.__query_status_api("all_status")

        if response is None or response.get("success") is False:
            raise RuntimeError(f"Request failed with: {response}")

        return response

    def get_device(self, id: Union[str, int]) -> dict:
        """Get the status of ONE device"""
        response = self.__query_status_api("status", id)

        if response is None or response.get("success") is False:
            raise RuntimeError(f"Request failed with: {response}")

        return response

    def get_device_ids(self) -> list[Union[str, int]]:
        """Get the ids of all devices"""
        response = self.get_all_devices()
        devices = response.get("data", {}).get("devices_status")

        return list(devices)  # Only get the ids of the devices

    def plug_s_turn_on(self, channel: int, device_id: str) -> list[Union[str, int]]:
        """Turn Plug S device on"""
        response = self.__set_parameter(
            "relay/control", {"channel": channel, "id": device_id, "turn": "on"}
        )

        return response

    def plug_s_turn_off(self, channel: int, device_id: str) -> list[Union[str, int]]:
        """Turn Plug S device off"""
        response = self.__set_parameter(
            "relay/control", {"channel": channel, "id": device_id, "turn": "off"}
        )

        return response



"""Test cases for the validation module"""
import pytest

from shellyapi.validations import validate_empty_string


def test_validation_ok() -> None:
    """Test if a non empty string is valid, thus returns None"""
    assert validate_empty_string("test", "hello") is None

def test_validation_error() -> None:
    """Test if an empty string raises an error"""
    with pytest.raises(Exception) as exception:
        validate_empty_string("", "hello")
    assert str(exception.value) == "hello must be specified"



"""Shelly validations module"""

def validate_empty_string(val: str, entity_name: str) -> None:
    """
    Validate (assert) if the string is empty or not.
    """

    assert isinstance(val, str) and val != "", f"{entity_name} must be specified"



"""Package setup"""

import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="shellyapi",
    version="0.2.0",
    author="",
    author_email="",
    description="A Python API for accessing the Shelly devices",
    license="",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: TBD License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    provides=["shellyapi"],
    install_requires=["requests"],
    setup_requires=["wheel"],
)



"""Test cases for the shelly api module"""
from unittest import mock
import pytest

from shellyapi.shellyapi import ShellyApi


DEVICE_REQUEST_RESPONSE = {
    "isok": True,
    "data": {
        "online": False,
        "device_status": {"total_power": -2500, "serial": 1000},
    },
}

DEVICES_REQUEST_RESPONSE = {
    "isok": True,
    "data": {
        "devices_status": {
            "bcff4dfd1c9b": {"total_power": -2500, "serial": 1000},
            "34945474b5ab": {"total_power": -3000, "serial": 1001},
            "4022d8891390": {
                "serial": 22000,
                "meters": [{"power": 10, "total": 30}],
                "relays": [{"ison": True}],
            },
        },
    },
}

DEVICES_IDS = ["bcff4dfd1c9b", "34945474b5ab", "4022d8891390"]

PLUG_S_TURN_RESPONSE = {"isok": True, "data": {"device_id": "4022d8891390"}}

# pylint: disable=unused-argument
def mocked_requests_get(*args, **kwargs):
    """Module handling mocked API requests"""

    # pylint: disable=too-few-public-methods
    class MockResponse:
        """Class handling mocked API responses"""

        def __init__(self, json_data, status_code):
            self.json_data = json_data
            self.status_code = status_code

        def json(self):
            """Return data as a JSON"""
            return self.json_data

    if args[0] == "http://localhost:3000/device/status":
        return MockResponse(DEVICE_REQUEST_RESPONSE, 200)

    if args[0] == "http://localhost:3000/device/all_status":
        return MockResponse(DEVICES_REQUEST_RESPONSE, 200)

    if args[0] == "http://localhost:3000/device/relay/control":
        return MockResponse(PLUG_S_TURN_RESPONSE, 200)

    if args[0] == "http://localhost:3001/device/all_status":
        return MockResponse(None, 200)

    return MockResponse(None, 404)


@mock.patch(
    "requests.post",
    mock.Mock(side_effect=mocked_requests_get),
)
def test_request_device() -> None:
    """Test if the device of the given ID is returned"""
    api = ShellyApi("http://localhost:3000", "TOKEN")
    assert api.get_device("bcff4dfd1c9b") == DEVICE_REQUEST_RESPONSE


@mock.patch(
    "requests.post",
    mock.Mock(side_effect=mocked_requests_get),
)
def test_request_all_devices() -> None:
    """Test if all devices are returned"""
    api = ShellyApi("http://localhost:3000", "TOKEN")
    assert api.get_all_devices() == DEVICES_REQUEST_RESPONSE


@mock.patch(
    "requests.post",
    mock.Mock(side_effect=mocked_requests_get),
)
def test_request_device_ids() -> None:
    """Test if all device ids are returned"""
    api = ShellyApi("http://localhost:3000", "TOKEN")
    assert api.get_device_ids() == DEVICES_IDS


@mock.patch(
    "requests.post",
    mock.Mock(side_effect=mocked_requests_get),
)
def test_request_turn_on() -> None:
    """Test if a device can be turned on"""
    api = ShellyApi("http://localhost:3000", "TOKEN")
    assert api.plug_s_turn_on(0, "4022d8891390") == PLUG_S_TURN_RESPONSE


@mock.patch(
    "requests.post",
    mock.Mock(side_effect=mocked_requests_get),
)
def test_request_turn_off() -> None:
    """Test if a device can be turned off"""
    api = ShellyApi("http://localhost:3000", "TOKEN")
    assert api.plug_s_turn_off(0, "4022d8891390") == PLUG_S_TURN_RESPONSE


@mock.patch(
    "requests.post",
    mock.Mock(side_effect=mocked_requests_get),
)
def test_request_status_error() -> None:
    """Test if API call fails"""
    api = ShellyApi("http://localhost:3001", "TOKEN")
    with pytest.raises(Exception) as exception:
        api.get_device_ids()
    assert str(exception.value) == "Request failed with: None"
